     ///////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM    //
   // Written by Andrew Broad                ///
  //                                        ////
 // AggInitFrame.java (1998:10:27 10:56)   /////
////////////////////////////////////////////////
// AggInitFrame is a subclass of Frame.    /////
//                                         /////
// Methods of AggInitFrame:                /////
// - static initialise                     /////
// - AggInitFrame(AggInit,Frame)           /////
// - AggInitFrame(AggInit,Frame,Hashtable) /////
// - static buildQueuedFrames              /////
// - buildFrame                            /////
// - getFrameClass                         /////
// - getNumber             [INHERITED]     /////
// - setNumber             [INHERITED]     /////
// - getSlotFiller(Slot)   [INHERITED]     /////
// - getSlotFiller(String) [INHERITED]     /////
// - setSlotFiller(Slot)   [INHERITED]     /////
// - setSlotFiller(String) [INHERITED]     /////
// - toString              [INHERITED]     /////
// - print                 [INHERITED]     /////
// - generateFIL           [INHERITED]     ////
// - static writeClass                     ///
// - static readClass                      //
////////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `AggInitFrame', a subclass of <a href="Frame.html">Frame</a>,
** represents a frame internally as an exjava AggInit object - such frames
** should only be used to represent aggregate initialisers in EXPRESS models.
*******************************************************************************/

public class AggInitFrame extends Frame {
	private static FrameClass _frameClass;
	private static boolean _initialised= false;
	private AggInit _underlyingAggInit;
	private static Vector/*AggInitFrame*/ _toBeBuilt;
	private static Hashtable/*AggInit,AggInitFrame*/ _cache;
	private Frame _theEntityOrType;
	
	/**********************************************************************/
	// static AggInitFrame.initialise : Express -> void
	/***********************************************************************
	** Initialises the AggInitFrame class by setting the class variable
	** _frameClass after looking up the `aggregate_initialiser' entity in
	** the given EXPRESS (meta)model.
	** @param metamodel An EXPRESS metamodel, which is expected to contain
	**  the entity `aggregate_initialiser'.
	***********************************************************************/

	public static void initialise (Express metamodel) {
		Entity e= metamodel.findEntity("aggregate_initialiser");
		if (e != null) {
			_frameClass= new FrameClass(e,metamodel);
		} else {
			System.err.println("Fatal error: failed to find "
			+ "ENTITY aggregate_initialiser in \"EXPRESS/"
			+ "frame_classes.exp\"!");
			System.exit(1);
		}
	}
	
	/**********************************************************************/
	// AggInitFrame constructor : AggInit ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava AggInit object
	** (it does so by copying the reference, which is maximally efficient
	** but the client should not then alter the AggInit directly, as that
	** would compromise the integrity of the frame).
	** @param instanceData An exjava AggInit object to wrap in this frame.
	** @param theEntity The entity which contains this aggregate initialiser.
 	** @exception FrameAlreadyExists If an AggInit has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public AggInitFrame (AggInit instanceData, Frame theEntity)
	                         throws FrameAlreadyExists
	{
		this(instanceData,theEntity,new Hashtable());
	}
	
	/**********************************************************************/
	// AggInitFrame constructor : (AggInit x Hashtable) ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava AggInit object
	** (it does so by copying the reference, which is maximally efficient,
	** but the client should not then alter the AggInit directly, as that
	** would compromise the integrity of the frame). This constructor allows
	** the aggregate_initialiser frame to be tagged with extra slots, given
	** in a hash table.
	** @param instanceData An exjava AggInit object to wrap in this frame.
	** @param theEntity The entity which contains this aggregate initialiser.
	** @param moreInstanceData Extra slots to tag this aggregate_initialiser
	**  frame with.
	** @exception FrameAlreadyExists If an AggInit has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public AggInitFrame (AggInit instanceData, Frame theEntity,
	                     Hashtable moreInstanceData)
	                    throws FrameAlreadyExists
	{
		super(moreInstanceData);

		if (_cache == null)
			_cache= new Hashtable();			

		_underlyingAggInit= instanceData;
		_theEntityOrType= theEntity;

		/////////////////////////////////////////////////////
		// Check that an AggInitFrame has not already been //
		// constructed for this AggInit.                   //
		/////////////////////////////////////////////////////
		if (_cache.containsKey(_underlyingAggInit))
			throw new FrameAlreadyExists((Frame)_cache.
				get(_underlyingAggInit));

		_cache.put(_underlyingAggInit,this);

		////////////////////////////////////////////////////////////
		// Build the frame now, provided its frame class has been //
		// constructed, otherwise queue it to be built later.     //
		////////////////////////////////////////////////////////////
		if (_initialised) {
			buildFrame();
			_frameClass.validateFrame(this);
			try {
				_frameClass.addInstance(this);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
				+ "wrong for an InvalidFrameClass to be thrown "
				+ "here!");
			}
			setNumber(_frameClass.getInstances().size());
		} else {
			if (_toBeBuilt == null)
				_toBeBuilt= new Vector();
			_toBeBuilt.addElement(this);
		}
	}
	
	/**********************************************************************/
	// static AggInitFrame.buildQueuedFrames : void -> void
	/***********************************************************************
	** Builds any AggInit Frames that are queued to be built (those that
	** were constructed before the frame class was constructed). Their slots
	** are filled from the exjava objects with which they were constructed,
	** and they are then validated.
	***********************************************************************/

	public static void buildQueuedFrames () {
		_initialised= true;
		if (_toBeBuilt == null)
			return;

		Enumeration toBeBuilt= _toBeBuilt.elements();
		while (toBeBuilt.hasMoreElements()) {
			AggInitFrame e= (AggInitFrame)toBeBuilt.nextElement();
			e.buildFrame();
			_frameClass.validateFrame(e);
			try {
				_frameClass.addInstance(e);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone "
				+ "badly wrong for an InvalidFrameClass"
				+ " to be thrown here!");
			}
			e.setNumber(_frameClass.getInstances().size());
		}
	}

	/**********************************************************************/
	// AggInitFrame.buildFrame : void -> void
	/***********************************************************************
	** Fills in the slots for this frame from the underlying exjava AggInit
	** object.
	***********************************************************************/

	public void buildFrame () {
		try {
			//////////////////////////////////////////////
			// elements : LIST [0 : ?]                  //
			//         OF aggregate_initialiser_element //
			//////////////////////////////////////////////
			setSlotFiller("elements", new OrderedVector
				(_underlyingAggInit.elems, _theEntityOrType));

		} catch (Exception x) {
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "AggInitFrame.buildFrame!");
		}
	}

	/**********************************************************************/
	// AggInitFrame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}
	
	/**********************************************************************/
	// static AggInitFrame.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of
	** AggInitFrame.
	** @param oo An output object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClass);
		oo.writeObject(_cache);
	}

	/**********************************************************************/
	// static AggInitFrame.readClass : ObjectInput -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** AggInitFrame.
	** @param ois An input object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClass= (FrameClass)oi.readObject();
			_cache= (Hashtable)oi.readObject();
			_initialised= true;
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the "
			+ "stupid compiler is just too dumb to realise this!");
		}
	}
}
