     /////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM  //
   // Written by Andrew Broad              ///
  //                                      ////
 // AggTypeFrame.java (1998:10:27 10:57) /////
//////////////////////////////////////////////
// AggTypeFrame is a subclass of Frame.  /////
//                                       /////
// Methods of AggTypeFrame:              /////
// - static initialise                   /////
// - AggTypeFrame(AggType)               /////
// - AggTypeFrame(AggType,Hashtable)     /////
// - static buildQueuedFrames            /////
// - buildFrame                          /////
// - getFrameClass                       /////
// - getNumber             [INHERITED]   /////
// - setNumber             [INHERITED]   /////
// - getSlotFiller(Slot)   [INHERITED]   /////
// - getSlotFiller(String) [INHERITED]   /////
// - setSlotFiller(Slot)   [INHERITED]   /////
// - setSlotFiller(String) [INHERITED]   /////
// - toString              [INHERITED]   /////
// - print                 [INHERITED]   /////
// - generateFIL           [INHERITED]   ////
// - static writeClass                   ///
// - static readClass                    //
//////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `AggTypeFrame', a subclass of <a href="Frame.html">Frame</a>,
** represents a frame internally as an exjava AggType object - such frames
** should only be used to represent aggregation types in EXPRESS models.
*******************************************************************************/

public class AggTypeFrame extends Frame {
	private static FrameClass _frameClass;
	private static boolean _initialised= false;
	private AggType _underlyingAggType;
	private static Vector/*AggTypeFrame*/ _toBeBuilt;
	private static Hashtable/*AggType,AggTypeFrame*/ _cache;
	
	/**********************************************************************/
	// static AggTypeFrame.initialise : Express -> void
	/***********************************************************************
	** Initialises the AggTypeFrame class by setting the class variable
	** _frameClass after looking up the `aggregation_type' entity in the
	** given EXPRESS (meta)model.
	** @param metamodel An EXPRESS metamodel, which is expected to contain
	**  the entity `aggregation_type'.
	***********************************************************************/

	public static void initialise (Express metamodel) {
		Entity e= metamodel.findEntity("aggregation_type");
		if (e != null) {
			_frameClass= new FrameClass(e,metamodel);
		} else {
			System.err.println("Fatal error: failed to find "
			+ "ENTITY aggregation_type in \"EXPRESS/"
			+ "frame_classes.exp\"!");
			System.exit(1);
		}
	}
	
	/**********************************************************************/
	// AggTypeFrame constructor : AggType ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava AggType object
	** (it does so by copying the reference, which is maximally efficient
	** but the client should not then alter the AggType directly, as that
	** would compromise the integrity of the frame).
	** @param instanceData An exjava AggType object to wrap in this frame.
	** @exception FrameAlreadyExists If an AggType has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public AggTypeFrame (AggType instanceData)
	                         throws FrameAlreadyExists
	{
		this(instanceData,new Hashtable());
	}
	
	/**********************************************************************/
	// AggTypeFrame constructor : (AggType x Hashtable) ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava AggType object
	** (it does so by copying the reference, which is maximally efficient,
	** but the client should not then alter the AggType directly, as that
	** would compromise the integrity of the frame). This constructor allows
	** the aggregation type frame to be tagged with extra slots, given in a
	** hash table.
	** @param instanceData An exjava AggType object to wrap in this frame.
	** @param moreInstanceData Extra slots to tag this aggregation type
	**  frame with.
	** @exception FrameAlreadyExists If an AggType has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public AggTypeFrame (AggType instanceData,
	                     Hashtable moreInstanceData)
	                    throws FrameAlreadyExists
	{
		super(moreInstanceData);

		if (_cache == null)
			_cache= new Hashtable();			

		_underlyingAggType= instanceData;

		/////////////////////////////////////////////////////
		// Check that an AggTypeFrame has not already been //
		// constructed for this AggType.                   //
		/////////////////////////////////////////////////////		
		if (_cache.containsKey(_underlyingAggType))
			throw new FrameAlreadyExists((Frame)_cache.
				get(_underlyingAggType));

		_cache.put(_underlyingAggType,this);

		////////////////////////////////////////////////////////////
		// Build the frame now, provided its frame class has been //
		// constructed, otherwise queue it to be built later.     //
		////////////////////////////////////////////////////////////
		if (_initialised) {
			buildFrame();
			_frameClass.validateFrame(this);
			try {
				_frameClass.addInstance(this);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
				+ "wrong for an InvalidFrameClass to be thrown "
				+ "here!");
			}
			setNumber(_frameClass.getInstances().size());
		} else {
			if (_toBeBuilt == null)
				_toBeBuilt= new Vector();
			_toBeBuilt.addElement(this);
		}
	}
	
	/**********************************************************************/
	// static AggTypeFrame.buildQueuedFrames : void -> void
	/***********************************************************************
	** Builds any AggType Frames that are queued to be built (those that
	** were constructed before the frame class was constructed). Their slots
	** are filled from the exjava objects with which they were constructed,
	** and they are then validated.
	***********************************************************************/

	public static void buildQueuedFrames () {
		_initialised= true;
		if (_toBeBuilt == null)
			return;

		Enumeration toBeBuilt= _toBeBuilt.elements();
		while (toBeBuilt.hasMoreElements()) {
			AggTypeFrame e= (AggTypeFrame)toBeBuilt.nextElement();
			e.buildFrame();
			_frameClass.validateFrame(e);
			try {
				_frameClass.addInstance(e);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
				+ "wrong for an InvalidFrameClass to be thrown "
				+ "here!");
			}
			e.setNumber(_frameClass.getInstances().size());
		}
	}

	/**********************************************************************/
	// AggTypeFrame.buildFrame : void -> void
	/***********************************************************************
	** Fills in the slots for this frame from the underlying exjava AggType
	** object.
	***********************************************************************/

	public void buildFrame () {
		try {
			/////////////////////////////
			// kind : aggregation_kind //
			/////////////////////////////
			String enum= "";
			switch (_underlyingAggType.kind) {
				case AggType.ARRAY: enum= "array_type"; break;
				case AggType.BAG:   enum= "bag_type";   break;
				case AggType.SET:   enum= "set_type";  break;
				case AggType.LIST:  enum= "list_type";  break;
			}
			setSlotFiller("kind", new EnumSymbol(enum));
	
			///////////////////////////////////
			// element_type : parameter_type //
			///////////////////////////////////
			setSlotFiller("element_type", Values.convert
				(_underlyingAggType.elemType));

			//////////////////////////////
			// lower_bound : expression //
			//////////////////////////////
			setSlotFiller("lower_bound", Values.convert
				(_underlyingAggType.lowerBound));

			//////////////////////////////
			// upper_bound : expression //
			//////////////////////////////
			setSlotFiller("upper_bound", Values.convert
				(_underlyingAggType.upperBound));

			/////////////////////////
			// is_unique : BOOLEAN //
			/////////////////////////
			setSlotFiller("is_unique",new Logical
				(_underlyingAggType.unique
				                   ? Logical.TRUE
				                   : Logical.FALSE));

			///////////////////////////
			// is_optional : BOOLEAN //
			///////////////////////////
			setSlotFiller("is_optional",new Logical
				(_underlyingAggType.optional
				                   ? Logical.TRUE
				                   : Logical.FALSE));

		} catch (Exception x) {
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "AggTypeFrame.buildFrame!");
		}
	}

	/**********************************************************************/
	// AggTypeFrame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}
	
	/**********************************************************************/
	// static AggTypeFrame.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of
	** AggTypeFrame.
	** @param oo An output object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClass);
		oo.writeObject(_cache);
	}

	/**********************************************************************/
	// static AggTypeFrame.readClass : ObjectInput -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** AggTypeFrame.
	** @param ois An input object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClass= (FrameClass)oi.readObject();
			_cache= (Hashtable)oi.readObject();
			_initialised= true;
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the "
			+ "stupid compiler is just too dumb to realise this!");
		}
	}
}
