     ///////////////////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM                //
   // Written by Andrew Broad                            ///
  //                                                    ////
 // BuiltInFnCallFrame.java (1998:10:27 11:00)         /////
////////////////////////////////////////////////////////////
// BuiltInFnCallFrame is a subclass of Frame.          /////
//                                                     /////
// Methods of BuiltInFnCallFrame:                      /////
// - static initialise                                 /////
// - BuiltInFnCallFrame(BuiltInFnCall,Frame)           /////
// - BuiltInFnCallFrame(BuiltInFnCall,Frame,Hashtable) /////
// - static buildQueuedFrames                          /////
// - buildFrame                                        /////
// - getFrameClass                                     /////
// - getNumber             [INHERITED]                 /////
// - setNumber             [INHERITED]                 /////
// - getSlotFiller(Slot)   [INHERITED]                 /////
// - getSlotFiller(String) [INHERITED]                 /////
// - setSlotFiller(Slot)   [INHERITED]                 /////
// - setSlotFiller(String) [INHERITED]                 /////
// - toString              [INHERITED]                 /////
// - print                 [INHERITED]                 /////
// - generateFIL           [INHERITED]                 ////
// - static writeClass                                 ///
// - static readClass                                  //
////////////////////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `BuiltInFnCallFrame', a subclass of <a href="Frame.html">Frame</a>,
** represents a frame internally as an exjava BuiltInFnCall object - such frames
** should only be used to represent built-in function calls in EXPRESS models.
*******************************************************************************/

public class BuiltInFnCallFrame extends Frame {
	private static FrameClass _frameClass;
	private static boolean _initialised= false;
	private BuiltInFnCall _underlyingBuiltInFnCall;
	private static Vector/*BuiltInFnCallFrame*/ _toBeBuilt;
	private static Hashtable/*BuiltInFnCall,BuiltInFnCallFrame*/ _cache;
	private Frame _theEntityOrType;
	
	/**********************************************************************/
	// static BuiltInFnCallFrame.initialise : Express -> void
	/***********************************************************************
	** Initialises the BuiltInFnCallFrame class by setting the class
	** variable _frameClass after looking up the `built_in_function_call'
	** entity in the given EXPRESS (meta)model.
	** @param metamodel An EXPRESS metamodel, which is expected to contain
	**  the entity `built_in_function_call'.
	***********************************************************************/

	public static void initialise (Express metamodel) {
		Entity e= metamodel.findEntity("built_in_function_call");
		if (e != null) {
			_frameClass= new FrameClass(e,metamodel);
		} else {
			System.err.println("Fatal error: failed to find "
			+ "ENTITY built_in_function_call in \"EXPRESS/"
			+ "frame_classes.exp\"!");
			System.exit(1);
		}
	}
	
	/**********************************************************************/
	// BuiltInFnCallFrame constructor : BuiltInFnCall ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava BuiltInFnCall object
	** (it does so by copying the reference, which is maximally efficient
	** but the client should not then alter the BuiltInFnCall directly, as that
	** would compromise the integrity of the frame).
	** @param instanceData An exjava BuiltInFnCall object to wrap in this frame.
	** @param theEntity The entity which contains this function call.
 	** @exception FrameAlreadyExists If an BuiltInFnCall has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public BuiltInFnCallFrame (BuiltInFnCall instanceData, Frame theEntity)
	                         throws FrameAlreadyExists
	{
		this(instanceData,theEntity,new Hashtable());
	}
	
	/**********************************************************************/
	// BuiltInFnCallFrame constructor : (BuiltInFnCall x Hashtable) ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava BuiltInFnCall object
	** (it does so by copying the reference, which is maximally efficient,
	** but the client should not then alter the BuiltInFnCall directly, as that
	** would compromise the integrity of the frame). This constructor allows
	** the function call frame to be tagged with extra slots, given in a
	** hash table.
	** @param instanceData An exjava BuiltInFnCall object to wrap in this frame.
	** @param theEntity The entity which contains this function call.
	** @param moreInstanceData Extra slots to tag this function call
	**  frame with.
	** @exception FrameAlreadyExists If an BuiltInFnCall has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public BuiltInFnCallFrame (BuiltInFnCall instanceData, Frame theEntity,
	                     Hashtable moreInstanceData)
	                    throws FrameAlreadyExists
	{
		super(moreInstanceData);

		if (_cache == null)
			_cache= new Hashtable();			

		_underlyingBuiltInFnCall= instanceData;
		_theEntityOrType= theEntity;

		///////////////////////////////////////////////////////////
		// Check that an BuiltInFnCallFrame has not already been //
		// constructed for this BuiltInFnCall.                   //
		///////////////////////////////////////////////////////////
		if (_cache.containsKey(_underlyingBuiltInFnCall))
			throw new FrameAlreadyExists((Frame)_cache.
				get(_underlyingBuiltInFnCall));

		_cache.put(_underlyingBuiltInFnCall,this);

		////////////////////////////////////////////////////////////
		// Build the frame now, provided its frame class has been //
		// constructed, otherwise queue it to be built later.     //
		////////////////////////////////////////////////////////////
		if (_initialised) {
			buildFrame();
			_frameClass.validateFrame(this);
			try {
				_frameClass.addInstance(this);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
				+ "wrong for an InvalidFrameClass to be thrown "
				+ "here!");
			}
			setNumber(_frameClass.getInstances().size());
		} else {
			if (_toBeBuilt == null)
				_toBeBuilt= new Vector();
			_toBeBuilt.addElement(this);
		}
	}
	
	/**********************************************************************/
	// static BuiltInFnCallFrame.buildQueuedFrames : void -> void
	/***********************************************************************
	** Builds any BuiltInFnCall Frames that are queued to be built (those that
	** were constructed before the frame class was constructed). Their slots
	** are filled from the exjava objects with which they were constructed,
	** and they are then validated.
	***********************************************************************/

	public static void buildQueuedFrames () {
		_initialised= true;
		if (_toBeBuilt == null)
			return;

		Enumeration toBeBuilt= _toBeBuilt.elements();
		while (toBeBuilt.hasMoreElements()) {
			BuiltInFnCallFrame e= (BuiltInFnCallFrame)toBeBuilt.nextElement();
			e.buildFrame();
			_frameClass.validateFrame(e);
			try {
				_frameClass.addInstance(e);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone "
				+ "badly wrong for an InvalidFrameClass"
				+ " to be thrown here!");
			}
			e.setNumber(_frameClass.getInstances().size());
		}
	}

	/**********************************************************************/
	// BuiltInFnCallFrame.buildFrame : void -> void
	/***********************************************************************
	** Fills in the slots for this frame from the underlying exjava BuiltInFnCall
	** object.
	***********************************************************************/

	public void buildFrame () {
		try {
			/////////////////////////////////////////
			// function_called : built_in_function //
			/////////////////////////////////////////
			String enum= "";
			switch (_underlyingBuiltInFnCall.kind) {
				case BuiltInFnCall.ABS: enum= "abs_function"; break;
				case BuiltInFnCall.ACOS: enum= "acos_function"; break;
				case BuiltInFnCall.ASIN: enum= "asin_function"; break;
				case BuiltInFnCall.ATAN: enum= "atan_function"; break;
				case BuiltInFnCall.BLENGTH: enum= "blength_function"; break;
				case BuiltInFnCall.COS: enum= "cos_function"; break;
				case BuiltInFnCall.EXISTS: enum= "exists_function"; break;
				case BuiltInFnCall.EXP: enum= "exp_function"; break;
				case BuiltInFnCall.FORMAT: enum= "format_function"; break;
				case BuiltInFnCall.HIBOUND: enum= "hibound_function"; break;
				case BuiltInFnCall.HIINDEX: enum= "hiindex_function"; break;
				case BuiltInFnCall.LENGTH: enum= "length_function"; break;
				case BuiltInFnCall.LOBOUND: enum= "lobound_function"; break;
				case BuiltInFnCall.LOG: enum= "log_function"; break;
				case BuiltInFnCall.LOG10: enum= "log10_function"; break;
				case BuiltInFnCall.LOG2: enum= "log2_function"; break;
				case BuiltInFnCall.LOINDEX: enum= "loindex_function"; break;
				case BuiltInFnCall.NVL: enum= "nvl_function"; break;
				case BuiltInFnCall.ODD: enum= "odd_function"; break;
				case BuiltInFnCall.ROLESOF: enum= "rolesof_function"; break;
				case BuiltInFnCall.SIN: enum= "sin_function"; break;
				case BuiltInFnCall.SIZEOF: enum= "sizeof_function"; break;
				case BuiltInFnCall.SQRT: enum= "sqrt_function"; break;
				case BuiltInFnCall.TAN: enum= "tan_function"; break;
				case BuiltInFnCall.TYPEOF: enum= "typeof_function"; break;
				case BuiltInFnCall.USEDIN: enum= "usedin_function"; break;
				case BuiltInFnCall.VALUE: enum= "value_function"; break;
				case BuiltInFnCall.VALUE_IN: enum= "value_in_function"; break;
				case BuiltInFnCall.VALUE_UNIQUE: enum= "value_unique_function"; break;
			}
			setSlotFiller("function_called", new EnumSymbol(enum));

			////////////////////////////////////////////////////
			// actual_parameters : LIST [0 : ?] OF expression //
			////////////////////////////////////////////////////
			setSlotFiller("actual_parameters", new OrderedVector
				(_underlyingBuiltInFnCall.params,
				 _theEntityOrType));
		} catch (Exception x) {
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "BuiltInFnCallFrame.buildFrame!");
		}
	}

	/**********************************************************************/
	// BuiltInFnCallFrame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}
	
	/**********************************************************************/
	// static BuiltInFnCallFrame.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of
	** BuiltInFnCallFrame.
	** @param oo An output object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClass);
		oo.writeObject(_cache);
	}

	/**********************************************************************/
	// static BuiltInFnCallFrame.readClass : ObjectInput -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** BuiltInFnCallFrame.
	** @param ois An input object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClass= (FrameClass)oi.readObject();
			_cache= (Hashtable)oi.readObject();
			_initialised= true;
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the "
			+ "stupid compiler is just too dumb to realise this!");
		}
	}
}
