     /////////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM      //
   // Written by Andrew Broad                  ///
  //                                          ////
 // DefinedTypeFrame.java (1998:11:03 11:00) /////
//////////////////////////////////////////////////
// DefinedTypeFrame is a subclass of Frame.  /////
//                                           /////
// Methods of DefinedTypeFrame:              /////
// - static initialise                       /////
// - DefinedTypeFrame(Type)                  /////
// - DefinedTypeFrame(Type,Hashtable)        /////
// - static buildQueuedFrames                /////
// - buildFrame                              /////
// - getFrameClass                           /////
// - getNumber             [INHERITED]       /////
// - setNumber             [INHERITED]       /////
// - getSlotFiller(Slot)   [INHERITED]       /////
// - getSlotFiller(String) [INHERITED]       /////
// - setSlotFiller(Slot)   [INHERITED]       /////
// - setSlotFiller(String) [INHERITED]       /////
// - toString              [INHERITED]       /////
// - print                 [INHERITED]       /////
// - generateFIL           [INHERITED]       /////
// - static writeClass                       /////
// - static readClass                        /////
// - getQueryVariables                       ////
// - pushQueryVariable                       ///
// - popQueryVariable                        //
//////////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `DefinedTypeFrame', a subclass of <a href="Frame.html">Frame</a>,
** represents a frame internally as an exjava Type object - such frames
** should only be used to represent defined types in EXPRESS models.
*******************************************************************************/

public class DefinedTypeFrame extends Frame {
	private static FrameClass _frameClass;
	private static boolean _initialised= false;
	private Type _underlyingType;
	private static Vector/*DefinedTypeFrame*/ _toBeBuilt;
	private static Hashtable/*Type,DefinedTypeFrame*/ _cache;
	private Vector/*Frame<query_variable>*/ _queryVariableStack;
	
	/**********************************************************************/
	// static DefinedTypeFrame.initialise : Express -> void
	/***********************************************************************
	** Initialises the DefinedTypeFrame class by setting the class variable
	** _frameClass after looking up the `defined_type' entity in the
	** given EXPRESS (meta)model.
	** @param metamodel An EXPRESS metamodel, which is expected to contain
	**  the entity `defined_type'.
	***********************************************************************/

	public static void initialise (Express metamodel) {
		Entity e= metamodel.findEntity("defined_type");
		if (e != null) {
			_frameClass= new FrameClass(e,metamodel);
		} else {
			System.err.println("Fatal error: failed to find "
			+ "ENTITY defined_type in \"EXPRESS/"
			+ "frame_classes.exp\"!");
			System.exit(1);
		}
	}
	
	/**********************************************************************/
	// DefinedTypeFrame constructor : Type ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava Type object
	** (it does so by copying the reference, which is maximally efficient
	** but the client should not then alter the Type directly, as that
	** would compromise the integrity of the frame).
	** @param instanceData An exjava Type object to wrap in this frame.
	** @exception FrameAlreadyExists If a Type has already been constructed
	**  for `instanceData'.
	***********************************************************************/

	public DefinedTypeFrame (Type instanceData)
	                         throws FrameAlreadyExists
	{
		this(instanceData,new Hashtable());
	}
	
	/**********************************************************************/
	// DefinedTypeFrame constructor : (Type x Hashtable) ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava Type object
	** (it does so by copying the reference, which is maximally efficient,
	** but the client should not then alter the Type directly, as that would
	** compromise the integrity of the frame). This constructor allows the
	** defined type frame to be tagged with extra slots, given in a hash
	** table.
	** @param instanceData An exjava Type object to wrap in this frame.
	** @param moreInstanceData Extra slots to tag this defined type frame
	**  with.
	** @exception FrameAlreadyExists If a Type has already been constructed
	**  for `instanceData'.
	***********************************************************************/

	public DefinedTypeFrame (Type instanceData,
	                         Hashtable moreInstanceData)
	                         throws FrameAlreadyExists
	{
		super(moreInstanceData);

		if (_cache == null)
			_cache= new Hashtable();			

		_underlyingType= instanceData;

		_queryVariableStack= new Vector ();

		/////////////////////////////////////////////////////////
		// Check that an DefinedTypeFrame has not already been //
		// constructed for this Type.                          //
		/////////////////////////////////////////////////////////		
		if (_cache.containsKey(_underlyingType))
			throw new FrameAlreadyExists((Frame)_cache.
				get(_underlyingType));

		_cache.put(_underlyingType,this);

//System.err.println("\t\tTYPE " + instanceData.id);

		////////////////////////////////////////////////////////////
		// Build the frame now, provided its frame class has been //
		// constructed, otherwise queue it to be built later.     //
		////////////////////////////////////////////////////////////
		if (_initialised) {
			buildFrame();
			_frameClass.validateFrame(this);
			try {
				_frameClass.addInstance(this);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
				+ "wrong for an InvalidFrameClass to be thrown "
				+ "here!");
			}
			setNumber(_frameClass.getInstances().size());

/*			try {
//System.out.println("\t\tEND_TYPE " + getSlotFiller("name") + "\n");
			} catch (InvalidSlot y) {
				System.err.println("DefinedTypeFrame.build" +
				"QueuedFrames: InvalidSlot");
				System.exit(1);
			}
*/		} else {
			if (_toBeBuilt == null)
				_toBeBuilt= new Vector();
			_toBeBuilt.addElement(this);
		}
	}
	
	/**********************************************************************/
	// static DefinedTypeFrame.buildQueuedFrames : void -> void
	/***********************************************************************
	** Builds any DefinedType Frames that are queued to be built (those that
	** were constructed before the frame class was constructed). Their slots
	** are filled from the exjava objects with which they were constructed,
	** and they are then validated.
	***********************************************************************/

	public static void buildQueuedFrames () {
		_initialised= true;
		if (_toBeBuilt == null)
			return;

		Enumeration toBeBuilt= _toBeBuilt.elements();
		while (toBeBuilt.hasMoreElements()) {
			DefinedTypeFrame e= (DefinedTypeFrame)toBeBuilt.
				            nextElement();
			e.buildFrame();
			_frameClass.validateFrame(e);
			try {
				_frameClass.addInstance(e);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone "
				+ "badly wrong for an InvalidFrameClass"
				+ " to be thrown here!");
			}
			e.setNumber(_frameClass.getInstances().size());

/*			try {
//System.out.println("\t\tEND_TYPE " + e.getSlotFiller("name") + "\n");
			} catch (InvalidSlot y) {
				System.err.println("DefinedTypeFrame.build" +
				"QueuedFrames: InvalidSlot");
				System.exit(1);
			}
*/		}
	}

	/**********************************************************************/
	// DefinedTypeFrame.buildFrame : void -> void
	/***********************************************************************
	** Fills in the slots for this frame from the underlying exjava Type
	** object.
	***********************************************************************/

	public void buildFrame () {
		try {
			///////////////////
			// name : STRING //
			///////////////////
			setSlotFiller("name",_underlyingType.id);

			////////////////////////////
			// tipe : underlying_type //
			////////////////////////////
			setSlotFiller("tipe", Values.convert
				(_underlyingType.type));
			
			////////////////////////////////////////////////////////
			// domain_rules : OPTIONAL SET [1 : ?] OF domain_rule //
			////////////////////////////////////////////////////////
			if (_underlyingType.rules.length > 0)
				setSlotFiller("domain_rules", new
				UnorderedVector(_underlyingType.rules,this));


		} catch (Exception x) {
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "DefinedTypeFrame.buildFrame!");
		}
	}

	/**********************************************************************/
	// DefinedTypeFrame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}
	
	/**********************************************************************/
	// static DefinedTypeFrame.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of
	** DefinedTypeFrame.
	** @param oo An output object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClass);
		oo.writeObject(_cache);
	}

	/**********************************************************************/
	// static DefinedTypeFrame.readClass : ObjectInput -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** DefinedTypeFrame.
	** @param ois An input object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClass= (FrameClass)oi.readObject();
			_cache= (Hashtable)oi.readObject();
			_initialised= true;
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the "
			+ "stupid compiler is just too dumb to realise this!");
		}
	}

	/**********************************************************************/
	// DefinedTypeFrame.getQueryVariables : void -> Vector
	/***********************************************************************
	** Returns the stack of query variables on this frame.
	** @return The query variables on this defined type frame.
	***********************************************************************/
 
	public Vector getQueryVariables () {
		return _queryVariableStack;
	}

	/**********************************************************************/
	// DefinedTypeFrame.pushQueryVariable : Frame -> void
	/***********************************************************************
	** Pushes a given query variable onto the stack.
	** @param queryVariable The query variable to be pushed onto this frame,
	**                      which is supposed to be a query_variable frame.
	***********************************************************************/
 
	public void pushQueryVariable (Frame queryVariable) {
		_queryVariableStack.addElement(queryVariable);
	}

	/**********************************************************************/
	// DefinedTypeFrame.popQueryVariable : void -> void
	/***********************************************************************
	** Pops the top query variable off of this frame.
	***********************************************************************/
 
	public void popQueryVariable () {
		_queryVariableStack.removeElementAt(0);
	}
}

