     ///////////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM        //
   // Written by Andrew Broad                    ///
  //                                            ////
 // ExplicitAttrFrame.java (1998:10:27 11:11)  /////
////////////////////////////////////////////////////
// ExplicitAttrFrame is a subclass of Frame.   /////
//                                             /////
// Methods of ExplicitAttrFrame:               /////
// - static initialise                         /////
// - ExplicitAttrFrame(ExplicitAttr)           /////
// - ExplicitAttrFrame(ExplicitAttr,Hashtable) /////
// - static buildQueuedFrames                  /////
// - buildFrame                                /////
// - getFrameClass                             /////
// - getNumber             [INHERITED]         /////
// - setNumber             [INHERITED]         /////
// - getSlotFiller(Slot)   [INHERITED]         /////
// - getSlotFiller(String) [INHERITED]         /////
// - setSlotFiller(Slot)   [INHERITED]         /////
// - setSlotFiller(String) [INHERITED]         /////
// - toString              [INHERITED]         /////
// - print                 [INHERITED]         /////
// - generateFIL           [INHERITED]         ////
// - static writeClass                         ///
// - static readClass                          //
////////////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `ExplicitAttrFrame', a subclass of <a href="Frame.html">Frame</a>,
** represents a frame internally as an exjava ExplicitAttr object - such frames
** should only be used to represent explicit attributes in EXPRESS models.
*******************************************************************************/

public class ExplicitAttrFrame extends Frame {
	private static FrameClass _frameClass;
	private static boolean _initialised= false;
	private ExplicitAttr _underlyingExplicitAttr;
	private static Vector/*ExplicitAttrFrame*/ _toBeBuilt;
	private static Hashtable/*ExplicitAttr,ExplicitAttrFrame*/ _cache;

	/**********************************************************************/
	// static ExplicitAttrFrame.initialise : Express -> void
	/***********************************************************************
	** Initialises the ExplicitAttrFrame class by setting the class variable
	** _frameClass after looking up the `explicit_attribute' entity in the
	** given EXPRESS (meta)model.
	** @param metamodel An EXPRESS metamodel, which is expected to contain
	**  the entity `explicit_attribute'.
	***********************************************************************/

	public static void initialise (Express metamodel) {
		Entity e= metamodel.findEntity("explicit_attribute");
		if (e != null) {
			_frameClass= new FrameClass(e,metamodel);
		} else {
			System.err.println("Fatal error: failed to find "
			+ "ENTITY explicit_attribute in \"EXPRESS/"
			+ "frame_classes.exp\"!");
			System.exit(1);
		}
	}
	
	/**********************************************************************/
	// ExplicitAttrFrame constructor : ExplicitAttr ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava ExplicitAttr
	** object (it does so by copying the reference, which is maximally
	** efficient, but the client should not then alter the ExplicitAttr
	** directly, as that would compromise the integrity of the frame).
	** @param instanceData An exjava ExplicitAttr object to wrap in this
	**                     frame.
	** @exception FrameAlreadyExists If an ExplicitAttrFrame has already
	**  been constructed for `instanceData'.
	***********************************************************************/

	public ExplicitAttrFrame (ExplicitAttr instanceData)
	                         throws FrameAlreadyExists
	{
		this(instanceData,new Hashtable());
	}
	
	/**********************************************************************/
	// ExplicitAttrFrame constructor :
	//	(ExplicitAttr x Hashtable) ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava ExplicitAttr
	** object (it does so by copying the reference, which is maximally
	** efficient, but the client should not then alter the ExplicitAttr
	** directly, as that would compromise the integrity of the frame). This
	** constructor allows the explicit attribute frame to be tagged with
	** extra slots, given in a hash table.
	** @param instanceData An exjava ExplicitAttr object to wrap in this
	**                     frame.
	** @param moreInstanceData Extra slots to tag this explicit attribute
	**  frame with.
	** @exception FrameAlreadyExists If an ExplicitAttrFrame has already
	**  been constructed for `instanceData'.
	***********************************************************************/

	public ExplicitAttrFrame (ExplicitAttr instanceData,
	                          Hashtable moreInstanceData)
	                         throws FrameAlreadyExists
	{
		super(moreInstanceData);

		if (_cache == null)
			_cache= new Hashtable();			

		_underlyingExplicitAttr= instanceData;

		//////////////////////////////////////////////////////////
		// Check that an ExplicitAttrFrame has not already been //
		// constructed for this ExplicitAttr.                   //
		//////////////////////////////////////////////////////////		
		if (_cache.containsKey(_underlyingExplicitAttr))
			throw new FrameAlreadyExists((Frame)_cache.
				get(_underlyingExplicitAttr));

		_cache.put(_underlyingExplicitAttr,this);

		////////////////////////////////////////////////////////////
		// Build the frame now, provided its frame class has been //
		// constructed, otherwise queue it to be built later.     //
		////////////////////////////////////////////////////////////
		if (_initialised) {
			buildFrame();
			_frameClass.validateFrame(this);
			try {
				_frameClass.addInstance(this);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
					+ "wrong for an InvalidFrameClass to be"
						+ " thrown here!");
			}
			setNumber(_frameClass.getInstances().size());
		} else {
			if (_toBeBuilt == null)
				_toBeBuilt= new Vector();
			_toBeBuilt.addElement(this);
		}
	}
	
	/**********************************************************************/
	// static ExplicitAttrFrame.buildQueuedFrames : void -> void
	/***********************************************************************
	** Builds any ExplicitAttrFrames that are queued to be built (those that
	** were constructed before the frame class was constructed). Their slots
	** are filled from the exjava objects with which they were constructed,
	** and they are then validated.
	***********************************************************************/

	public static void buildQueuedFrames () {
		_initialised= true;
		if (_toBeBuilt == null)
			return;

		Enumeration toBeBuilt= _toBeBuilt.elements();
		while (toBeBuilt.hasMoreElements()) {
			ExplicitAttrFrame e= (ExplicitAttrFrame)
				toBeBuilt.nextElement();
			e.buildFrame();
			_frameClass.validateFrame(e);
			try {
				_frameClass.addInstance(e);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone "
				+ "badly wrong for an InvalidFrameClass"
				+ " to be thrown here!");
			}
			e.setNumber(_frameClass.getInstances().size());
		}
	}

	/**********************************************************************/
	// ExplicitAttrFrame.buildFrame : void -> void
	/***********************************************************************
	** Fills in the slots for this frame from the underlying exjava
	** ExplicitAttr object.
	***********************************************************************/

	public void buildFrame () {
		try {
			///////////////////
			// name : STRING //
			///////////////////
			setSlotFiller("name",_underlyingExplicitAttr.id);

			///////////////////////////
			// tipe : parameter_type //
			///////////////////////////
			setSlotFiller("tipe", Values.convert
				(_underlyingExplicitAttr.type));

			//////////////////////////////////
			// owner : OPTIONAL entity_type //
			//////////////////////////////////
			setSlotFiller("owner", Values.convert
				(_underlyingExplicitAttr.owner));
			
			///////////////////////////
			// is_optional : BOOLEAN //
			///////////////////////////
			setSlotFiller("is_optional", new Logical
				(_underlyingExplicitAttr.optional
				                        ? Logical.TRUE
				                        : Logical.FALSE));

		} catch (Exception x) {
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "ExplicitAttrFrame.buildFrame!");
		}
	}

	/**********************************************************************/
	// ExplicitAttrFrame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}
	
	/**********************************************************************/
	// static ExplicitAttrFrame.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of
	** ExplicitAttrFrame.
	** @param oo An output object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClass);
		oo.writeObject(_cache);
	}

	/**********************************************************************/
	// static ExplicitAttrFrame.readClass : ObjectInput -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** ExplicitAttrFrame.
	** @param ois An input object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClass= (FrameClass)oi.readObject();
			_cache= (Hashtable)oi.readObject();
			_initialised= true;
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the "
			+ "stupid compiler is just too dumb to realise this!");
		}
	}
}
