     /////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM ///
   // Written by Andrew Broad             ////
  //                                     /////
 // FrameClass.java (1998:10:27 11:23)  //////
//////////////////////////////////////////////
// Methods of FrameClass:              ///////
// - FrameClass constructor            ///////
// - getClassFrame                     ///////
// - getName                           ///////
// - getSuperClasses                   ///////
// - subsumes                          ///////
// - getSlots                          ///////
// - validateFrame                     ///////
// - getInstances                      ///////
// - addInstance                       ///////
// - toString                          ///////
// - getSlot                           ///////
// - generateFILforInstances           ///////
// - static getAllFrameClasses         ///////
// - static lookup(String)             //////
// - static lookup(String,Express)     /////
// - static lookup(Entity,Express)     ////
// - static writeClass                 ///
// - static readClass                  //
////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `FrameClass' represents a class of frames, serving to classify them
** and to specify the slots they should have.
*******************************************************************************/

public class FrameClass implements Serializable {
	private Frame _classFrame;
	private String _name;
	private Vector/*Frame*/ _instances;
	private Vector/*FrameClass*/ _superClasses;
	private Vector/*Slot*/ _slots;
	private Hashtable/*String,Slot*/ _slotIndex;
	private static Hashtable/*String,FrameClass*/ _frameClassIndex;

        /**********************************************************************/
        // FrameClass constructor : Entity ->
        /***********************************************************************
        ** Constructs this frame class by plugging in the given exjava Entity
        ** object which is taken to represent it (it does so by copying the
        ** reference, which is maximally efficient, but the client should not
        ** then alter the Entity directly, as that would compromise the
        ** integrity of the frame).
        ** @param classFrame An exjava Entity object to wrap in this frame.
	** @param frameClassModel The EXPRESS model in which this frame class
	**  is defined, in case it has any superclasses which haven't been
	**  constructed yet.
        ***********************************************************************/

	public FrameClass (Entity classFrame, Express frameClassModel) {
		Slot slot; FrameClass frameClass;

		_name= classFrame.id;		
		_instances= new Vector();
		_superClasses= new Vector();
		_slots= new Vector();
		_slotIndex= new Hashtable();
		if (_frameClassIndex == null)
			_frameClassIndex= new Hashtable();
					
		try {
			_classFrame= new EntityFrame(classFrame);
		} catch (FrameAlreadyExists y) {
			_classFrame= y._frame;
		}

		////////////////////////////////////////
		// Construct the slots inherited from //
		// this frame class's superclasses    //
		////////////////////////////////////////
		for (int s= 0; s < classFrame.subtype.length; s++) {
			frameClass= null;
			try {
				frameClass= lookup(classFrame.subtype[s].id,
					frameClassModel);
				_superClasses.addElement(frameClass);
				Vector inheritedSlots= frameClass.getSlots();
				for (int a= 0; a < inheritedSlots.size(); a++) {
					slot= (Slot)inheritedSlots.elementAt(a);
					_slots.addElement(slot);
					_slotIndex.put(slot.getName(),slot);
				}
			} catch (FrameClassDoesNotExist x) {
				System.err.println("Something has gone very "
				+ "wrong for the frame class " + classFrame.
				subtype[s].id + " not to exist!");
			}
		}

		/////////////////////////////////////////////////
		// Construct this frame class's explicit slots //
		/////////////////////////////////////////////////
		for (int e= 0; e < classFrame.explicits.length; e++) {
			slot= new Slot(classFrame.explicits[e],this,
			               frameClassModel);
			_slots.addElement(slot);
			_slotIndex.put(classFrame.explicits[e].id,slot);
		}

		////////////////////////////////////////////////
		// Construct this frame class's derived slots //
		////////////////////////////////////////////////
		for (int d= 0; d < classFrame.deriveds.length; d++) {
			slot= new Slot(classFrame.deriveds[d],this,
			               frameClassModel);
			_slots.addElement(slot);
			_slotIndex.put(classFrame.deriveds[d].id,slot);
		}

		////////////////////////////////////////////////
		// Construct this frame class's inverse slots //
		////////////////////////////////////////////////
		for (int i= 0; i < classFrame.inverses.length; i++) {
			slot= new Slot(classFrame.inverses[i],this,
			               frameClassModel);
			_slots.addElement(slot);
			_slotIndex.put(classFrame.inverses[i].id,slot);
		}

		_frameClassIndex.put(classFrame.id,this);
	}

	/**********************************************************************/
	// FrameClass.getClassFrame : void -> Frame
	/***********************************************************************
	** Delivers up the underlying frame of this frame class.
	** @return The class frame representing this frame class.
	***********************************************************************/
 
	public Frame getClassFrame () {
		return _classFrame;
	}
	
	/**********************************************************************/
	// FrameClass.getName : void -> String
	/***********************************************************************
	** Returns the name of this frame class.
	** @return The name of this frame class.
	***********************************************************************/
 	
	public String getName () {
		return _name;
	}
	
	/**********************************************************************/
	// FrameClass.getSuperClasses : void -> Vector
	/***********************************************************************
	** Delivers up this frame class' superclasses.
	** @return A vector of this frame class' superclasses.
	***********************************************************************/
 
	public Vector getSuperClasses () {
		return _superClasses;
	}
	
        /**********************************************************************/
        // FrameClass.subsumes : FrameClass -> boolean
        /***********************************************************************
        ** Does this frame class subsume another? Returns true just in case the
        ** given frame class is equal to or a subclass of this one.
        ** @param subClass The frame class to test whether it is a subclass of
        **  this one.
        ** @return Whether this frame class does subsume `subClass'.
        ***********************************************************************/
 
        public boolean subsumes (FrameClass subClass) {
		if (subClass == this)
			return true;
		
		//////////////////////////////////////////////////
		// Move up the hierarchy from subClass          //
		// and see if you collide with this frame class //
		//////////////////////////////////////////////////
		for (int s= 0; s < subClass.getSuperClasses().size(); s++) {
			if (subsumes((FrameClass)subClass.getSuperClasses().
			elementAt(s))) {
				return true;
			}
		}
		return false;
	}
	
	/**********************************************************************/
	// FrameClass.getSlots : void -> Vector
	/***********************************************************************
	** Delivers up the slots in this frame.
	** @return A vector of the slots in this frame.
	***********************************************************************/
 
	public Vector getSlots () {
		return _slots;
	}

	/**********************************************************************/
	// FrameClass.validateFrame : Frame -> void
	/***********************************************************************
	** Checks that the given frame is a valid instance of this frame class,
	** raising exceptions if it is not. Validation isn't actually
	** implemented - this is just a hook.
	** @param instance The frame to validate.
	***********************************************************************/
 
 	public void validateFrame (Frame instance)
	                          /* throws InvalidSlotFiller */
	{	//Slot slot;
	
/*		for (int s= 0; s < getSlots().size(); s++) {
			slot= (Slot)getSlots().elementAt(s);
			if (! slot.validFiller(instance.getSlotFiller (slot)))
				throw new InvalidSlotFiller();
		}
*/	}
	
	/**********************************************************************/
	// FrameClass.getInstances : void -> Vector
	/***********************************************************************
	** Returns the instances of this frame class that have been
	** created in the program's lifetime.
	** @return A vector of frames that are instances of this frame class.
	***********************************************************************/
 
	public Vector getInstances () {
		return _instances;
	}
	
	/**********************************************************************/
	// FrameClass.addInstance : Frame -> void
	/***********************************************************************
	** Adds the given frame to the list of instances of this FrameClass.
	** @param frame The instance frame to add (must be of this frame class).
	** @exception InvalidFrameClass If the given frame is not of this frame
	**  class!
	***********************************************************************/
 
	public void addInstance (Frame frame)
	                        throws InvalidFrameClass
	{
		if (frame.getFrameClass() != this)
			throw new InvalidFrameClass();

		_instances.addElement(frame);
	}
	
	/**********************************************************************/
	// FrameClass.toString : void -> String
	/***********************************************************************
	** Returns the name of this frame class.
	** @returns The string representation of this frame class.
	***********************************************************************/

	public String toString () {
		try {
			return (String)_classFrame.
				getSlotFiller("name");
		} catch (InvalidSlot x) {
			System.err.println("Something has gone "
			+ "horribly wrong for FrameClass."
			+ "toString to throw an InvalidSlot!");
			System.exit(1); return "";
		}
	}

        /**********************************************************************/
        // FrameClass.getSlot : String -> Slot
        /***********************************************************************
        ** Looks up the slot with the given name.
        ** @param slotName The name of the slot to return.
        ** @return The Slot object whose name is `slotName'.
        ** @exception InvalidSlot If the frame class does not have a slot of
        **                        that name.
        ***********************************************************************/
 
        public Slot getSlot (String slotName)
                            throws InvalidSlot
        {
		Slot slot= (Slot)_slotIndex.get(slotName);
		if (slot == null)
			throw new InvalidSlot();
		return slot;
        }

	/**********************************************************************/
	// Frame.generateFILforInstances : boolean -> void
	/***********************************************************************
	** Writes all instances of this frame class into an HTML document,
	** called <TT><<I>frame-class</I>>.html</TT>.
	** @param HTML If true, file will be augmented with HTML tags.
	***********************************************************************/

	public void generateFILforInstances (boolean HTML) {
		try {
			String filename= (String)_classFrame.
				getSlotFiller("name");
			if (HTML) filename+= ".html";
			else      filename+= ".fil";

			System.out.println("Generating " + filename);
			FileOutputStream fo= new FileOutputStream
			                       ("/usr/local/tmp/users/broada/"
			                       + filename);
			PrintWriter pw= new PrintWriter(fo,true);

			if (HTML) pw.println("<HTML>");
			if (HTML) pw.println("<H1>" + 
				  _classFrame.getSlotFiller("name") + "</H1>");

                	Enumeration instanceFrames= getInstances().elements();
                	while (instanceFrames.hasMoreElements()) {
                		if (HTML) pw.println("<HR>");
                		else      pw.println();
                       		((Frame)instanceFrames.nextElement()).
                       			generateFIL(pw,HTML);
			}
			if (HTML) pw.println("</HTML>");
		} catch (InvalidSlot x) {
			System.err.println("Something has gone "
			+ "horribly wrong for FrameClass."
			+ "generateFILforInstances to throw an InvalidSlot!");
			System.exit(1);
		} catch (IOException y) {
			System.err.println(y);
			System.exit(1);
		}
	}

        /**********************************************************************/
        // static FrameClass.getAllFrameClasses : void -> Enumeration
        /***********************************************************************
        ** Returns all the frame classes in the system.
        ** @return An enumeration of all the frame classes in the system.
        ***********************************************************************/

	public static Enumeration getAllFrameClasses () {
		return _frameClassIndex.elements();
	}

        /**********************************************************************/
        // static FrameClass.lookup : String -> FrameClass
        /***********************************************************************
        ** Looks up the frame class with the given name.
        ** @param frameClassName The name of the frame class to return.
        ** @return The FrameClass object whose name is `frameClassName'.
        ** @exception FrameClassDoesNotExist If there is not a frame class of
        **  that name.
        ***********************************************************************/
 
        public static FrameClass lookup (String frameClassName)
                            throws FrameClassDoesNotExist
        {
        	FrameClass frameClass= (FrameClass)_frameClassIndex.
        		get(frameClassName);
        	if (frameClass == null)
			throw new FrameClassDoesNotExist();

        	return frameClass;
        }

        /**********************************************************************/
        // static FrameClass.lookup : String x Express -> FrameClass
        /***********************************************************************
        ** Looks up the frame class with the given name.
        ** @param frameClassName The name of the frame class to return.
        ** @param frameClassModel The EXPRESS model in which `frameClassName'
        **  is defined, in case that frame class hasn't been constructed yet.
        ** @return The FrameClass object whose name is `frameClassName'.
        ** @exception FrameClassDoesNotExist If there is not a frame class of
        **  that name.
        ***********************************************************************/
 
        public static FrameClass lookup (String frameClassName,
                                         Express frameClassModel)
                            throws FrameClassDoesNotExist
        {
        	FrameClass frameClass= (FrameClass)_frameClassIndex.
        		get(frameClassName);
        	if (frameClass == null) {
        		Entity e= frameClassModel.findEntity(frameClassName);
        		if (e != null)
        			frameClass= new FrameClass(e,frameClassModel);
        		else
        			throw new FrameClassDoesNotExist();
        	}
        	return frameClass;
        }

        /**********************************************************************/
        // static FrameClass.lookup : Entity x Express -> FrameClass
        /***********************************************************************
        ** Looks up the frame class with name of the given entity.
        ** @param frameClassEntity The entity of the frame class to return.
        ** @param frameClassModel The EXPRESS model in which `frameClassName'
        **  is defined, in case that frame class hasn't been constructed yet.
        ** @return The FrameClass object constructed from `frameClassEntity'.
        ***********************************************************************/
 
        public static FrameClass lookup (Entity frameClassEntity,
                                         Express frameClassModel)
        {
        	FrameClass frameClass= (FrameClass)_frameClassIndex.
        		                           get(frameClassEntity.id);
        	if (frameClass == null)
       			frameClass= new FrameClass(frameClassEntity,
       			                           frameClassModel);
        	return frameClass;
        }

	/**********************************************************************/
	// static FrameClass.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of FrameClass.
	** @param oo An output stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClassIndex);
	}

	/**********************************************************************/
	// static FrameClass.readClass : ObjectInputStream -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** FrameClass.
	** @param oi An input stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClassIndex= (Hashtable)oi.readObject();
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the"
			+ "stupid compiler is just too dumb to realise this!");
		}
	}
}
