     /////////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM      //
   // Written by Andrew Broad                  ///
  //                                          ////
 // InverseAttrFrame.java (1998:11:25 10:42) /////
//////////////////////////////////////////////////
// InverseAttrFrame is a subclass of Frame.  /////
//                                           /////
// Methods of InverseAttrFrame:              /////
// - static initialise                       /////
// - InverseAttrFrame(InverseAttr)           /////
// - InverseAttrFrame(InverseAttr,Hashtable) /////
// - static buildQueuedFrames                /////
// - buildFrame                              /////
// - getFrameClass                           /////
// - getNumber             [INHERITED]       /////
// - setNumber             [INHERITED]       /////
// - getSlotFiller(Slot)   [INHERITED]       /////
// - getSlotFiller(String) [INHERITED]       /////
// - setSlotFiller(Slot)   [INHERITED]       /////
// - setSlotFiller(String) [INHERITED]       /////
// - toString              [INHERITED]       /////
// - print                 [INHERITED]       /////
// - generateFIL           [INHERITED]       ////
// - static writeClass                       ///
// - static readClass                        //
//////////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `InverseAttrFrame', a subclass of <a href="Frame.html">Frame</a>,
** represents a frame internally as an exjava InverseAttr object - such frames
** should only be used to represent inverse attributes in EXPRESS models.
*******************************************************************************/

public class InverseAttrFrame extends Frame {
	private static FrameClass _frameClass;
	private static boolean _initialised= false;
	private InverseAttr _underlyingInverseAttr;
	private static Vector/*InversetAttrFrame*/ _toBeBuilt;
	private static Hashtable/*InverseAttr,InverseAttrFrame*/ _cache;
	
	/**********************************************************************/
	// static InverseAttrFrame.initialise : Express -> void
	/***********************************************************************
	** Initialises the InverseAttrFrame class by setting the class variable
	** _frameClass after looking up the `inverse_attribute' entity in the
	** given EXPRESS (meta)model.
	** @param metamodel An EXPRESS metamodel, which is expected to contain
	**  the entity `inverse_attribute'.
	***********************************************************************/

	public static void initialise (Express metamodel) {
		Entity e= metamodel.findEntity("inverse_attribute");
		if (e != null) {
			_frameClass= new FrameClass(e,metamodel);
		} else {
			System.err.println("Fatal error: failed to find "
			+ "ENTITY inverse_attribute in \"EXPRESS/"
			+ "frame_classes.exp\"!");
			System.exit(1);
		}
	}
	
	/**********************************************************************/
	// InverseAttrFrame constructor : InverseAttr ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava InverseAttr
	** object (it does so by copying the reference, which is maximally
	** efficient, but the client should not then alter the InverseAttr
	** directly, as that would compromise the integrity of the frame).
	** @param instanceData An exjava InverseAttr object to wrap in this
	**                     frame.
	** @exception FrameAlreadyExists If an InverseAttrFrame has already
	**  been constructed for `instanceData'.
	***********************************************************************/

	public InverseAttrFrame (InverseAttr instanceData)
	                         throws FrameAlreadyExists
	{
		this(instanceData,new Hashtable());
	}
	
	/**********************************************************************/
	// InverseAttrFrame constructor :
	//	(InverseAttr x Hashtable) ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava InverseAttr
	** object (it does so by copying the reference, which is maximally
	** efficient, but the client should not then alter the InverseAttr
	** directly, as that would compromise the integrity of the frame). This
	** constructor allows the inverse attribute frame to be tagged with
	** extra slots, given in a hash table.
	** @param instanceData An exjava InverseAttr object to wrap in this
	**                     frame.
	** @param moreInstanceData Extra slots to tag this inverse attribute
	**  frame with.
	** @exception FrameAlreadyExists If an InverseAttrFrame has already
	**  been constructed for `instanceData'.
	***********************************************************************/

	public InverseAttrFrame (InverseAttr instanceData,
	                         Hashtable moreInstanceData)
	                         throws FrameAlreadyExists
	{
		super(moreInstanceData);

		if (_cache == null)
			_cache= new Hashtable();			

		_underlyingInverseAttr= instanceData;

		/////////////////////////////////////////////////////////
		// Check that an InverseAttrFrame has not already been //
		// constructed for this InverseAttr.                   //
		/////////////////////////////////////////////////////////		
		if (_cache.containsKey(_underlyingInverseAttr))
			throw new FrameAlreadyExists((Frame)_cache.
				get(_underlyingInverseAttr));

		_cache.put(_underlyingInverseAttr,this);

		////////////////////////////////////////////////////////////
		// Build the frame now, provided its frame class has been //
		// constructed, otherwise queue it to be built later.     //
		////////////////////////////////////////////////////////////
		if (_initialised) {
			buildFrame();
			_frameClass.validateFrame(this);
			try {
				_frameClass.addInstance(this);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
					+ "wrong for an InvalidFrameClass to be"
						+ " thrown here!");
			}
			setNumber(_frameClass.getInstances().size());
		} else {
			if (_toBeBuilt == null)
				_toBeBuilt= new Vector();
			_toBeBuilt.addElement(this);
		}
	}
	
	/**********************************************************************/
	// static InverseAttrFrame.buildQueuedFrames : void -> void
	/***********************************************************************
	** Builds any InverseAttrFrames that are queued to be built (those that
	** were constructed before the frame class was constructed). Their slots
	** are filled from the exjava objects with which they were constructed,
	** and they are then validated.
	***********************************************************************/

	public static void buildQueuedFrames () {
		_initialised= true;
		if (_toBeBuilt == null)
			return;

		Enumeration toBeBuilt= _toBeBuilt.elements();
		while (toBeBuilt.hasMoreElements()) {
			InverseAttrFrame e= (InverseAttrFrame)
				toBeBuilt.nextElement();
			e.buildFrame();
			_frameClass.validateFrame(e);
			try {
				_frameClass.addInstance(e);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone "
				+ "badly wrong for an InvalidFrameClass"
				+ " to be thrown here!");
			}
			e.setNumber(_frameClass.getInstances().size());
		}
	}

	/**********************************************************************/
	// InverseAttrFrame.buildFrame : void -> void
	/***********************************************************************
	** Fills in the slots for this frame from the underlying exjava
	** InverseAttr object.
	***********************************************************************/

	public void buildFrame () {
		try {
			///////////////////
			// name : STRING //
			///////////////////
			setSlotFiller("name",_underlyingInverseAttr.id);

			///////////////////////////
			// tipe : parameter_type //
			///////////////////////////
			setSlotFiller("tipe", Values.convert
				(_underlyingInverseAttr.type));

			//////////////////////////////////
			// owner : OPTIONAL entity_type //
			//////////////////////////////////
			setSlotFiller("owner", Values.convert
				(_underlyingInverseAttr.owner));

			///////////////////////////////
			// for_attribute : attribute //
			///////////////////////////////
			setSlotFiller("for_attribute", Values.convert
				(_underlyingInverseAttr.getForRef()));

		} catch (Exception x) {
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "InverseAttrFrame.buildFrame!");
		}
	}

	/**********************************************************************/
	// InverseAttrFrame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}
	
	/**********************************************************************/
	// static InverseAttrFrame.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of
	** InverseAttrFrame.
	** @param oo An output object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClass);
		oo.writeObject(_cache);
	}

	/**********************************************************************/
	// static InverseAttrFrame.readClass : ObjectInput -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** InverseAttrFrame.
	** @param ois An input object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClass= (FrameClass)oi.readObject();
			_cache= (Hashtable)oi.readObject();
			_initialised= true;
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the "
			+ "stupid compiler is just too dumb to realise this!");
		}
	}
}
