     /////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM  //
   // Written by Andrew Broad              ///
  //                                      ////
 // LiteralFrame.java (1998:10:27 11:29) /////
//////////////////////////////////////////////
// LiteralFrame is a subclass of Frame.  /////
//                                       /////
// Methods of LiteralFrame:              /////
// - static initialise                   /////
// - LiteralFrame(Literal)               /////
// - LiteralFrame(Literal,Hashtable)     /////
// - static buildQueuedFrames            /////
// - buildFrame                          /////
// - getFrameClass                       /////
// - getNumber             [INHERITED]   /////
// - setNumber             [INHERITED]   /////
// - getSlotFiller(Slot)   [INHERITED]   /////
// - getSlotFiller(String) [INHERITED]   /////
// - setSlotFiller(Slot)   [INHERITED]   /////
// - setSlotFiller(String) [INHERITED]   /////
// - toString              [INHERITED]   /////
// - print                 [INHERITED]   /////
// - generateFIL           [INHERITED]   ////
// - static writeClass                   ///
// - static readClass                    //
//////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `LiteralFrame', a subclass of <a href="Frame.html">Frame</a>,
** represents a frame internally as an exjava Literal object - such frames
** should only be used to represent literals in EXPRESS models.
*******************************************************************************/

public class LiteralFrame extends Frame {
	private static FrameClass _frameClass;
	private static boolean _initialised= false;
	private Literal _underlyingLiteral;
	private static Vector/*LiteralFrame*/ _toBeBuilt;
	private static Hashtable/*Literal,LiteralFrame*/ _cache;
	
	/**********************************************************************/
	// static LiteralFrame.initialise : Express -> void
	/***********************************************************************
	** Initialises the LiteralFrame class by setting the class variable
	** _frameClass after looking up the `literal_expression' entity in the
	** given EXPRESS (meta)model.
	** @param metamodel An EXPRESS metamodel, which is expected to contain
	**  the entity `literal_expression'.
	***********************************************************************/

	public static void initialise (Express metamodel) {
		Entity e= metamodel.findEntity("literal_expression");
		if (e != null) {
			_frameClass= new FrameClass(e,metamodel);
		} else {
			System.err.println("Fatal error: failed to find "
			+ "ENTITY literal_expression in \"EXPRESS/"
			+ "frame_classes.exp\"!");
			System.exit(1);
		}
	}
	
	/**********************************************************************/
	// LiteralFrame constructor : Literal ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava Literal
	** object (it does so by copying the reference, which is maximally
	** efficient, but the client should not then alter the Literal
	** directly, as that would compromise the integrity of the frame).
	** @param instanceData An exjava Literal object to wrap in this
	**                     frame.
	** @exception FrameAlreadyExists If an LiteralFrame has already
	**  been constructed for `instanceData'.
	***********************************************************************/

	public LiteralFrame (Literal instanceData)
	                    throws FrameAlreadyExists
	{
		this(instanceData,new Hashtable());
	}
	
	/**********************************************************************/
	// LiteralFrame constructor :
	//	(Literal x Hashtable) ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava Literal
	** object (it does so by copying the reference, which is maximally
	** efficient, but the client should not then alter the Literal
	** directly, as that would compromise the integrity of the frame). This
	** constructor allows the literal frame to be tagged with extra slots,
	** given in a hash table.
	** @param instanceData An exjava Literal object to wrap in this frame.
	** @param moreInstanceData Extra slots to tag this literal frame with.
	** @exception FrameAlreadyExists If an LiteralFrame has already
	**  been constructed for `instanceData'.
	***********************************************************************/

	public LiteralFrame (Literal instanceData,
	                     Hashtable moreInstanceData)
	                    throws FrameAlreadyExists
	{
		super(moreInstanceData);

		if (_cache == null)
			_cache= new Hashtable();			

		_underlyingLiteral= instanceData;

		/////////////////////////////////////////////////////
		// Check that an LiteralFrame has not already been //
		// constructed for this Literal.                   //
		/////////////////////////////////////////////////////		
		if (_cache.containsKey(_underlyingLiteral))
			throw new FrameAlreadyExists((Frame)_cache.
				get(_underlyingLiteral));

		_cache.put(_underlyingLiteral,this);

		////////////////////////////////////////////////////////////
		// Build the frame now, provided its frame class has been //
		// constructed, otherwise queue it to be built later.     //
		////////////////////////////////////////////////////////////
		if (_initialised) {
			buildFrame();
			_frameClass.validateFrame(this);
			try {
				_frameClass.addInstance(this);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
					+ "wrong for an InvalidFrameClass to be"
						+ " thrown here!");
			}
			setNumber(_frameClass.getInstances().size());
		} else {
			if (_toBeBuilt == null)
				_toBeBuilt= new Vector();
			_toBeBuilt.addElement(this);
		}
	}
	
	/**********************************************************************/
	// static LiteralFrame.buildQueuedFrames : void -> void
	/***********************************************************************
	** Builds any LiteralFrames that are queued to be built (those that
	** were constructed before the frame class was constructed). Their slots
	** are filled from the exjava objects with which they were constructed,
	** and they are then validated.
	***********************************************************************/

	public static void buildQueuedFrames () {
		_initialised= true;
		if (_toBeBuilt == null)
			return;

		Enumeration toBeBuilt= _toBeBuilt.elements();
		while (toBeBuilt.hasMoreElements()) {
			LiteralFrame e= (LiteralFrame)toBeBuilt.nextElement();
			e.buildFrame();
			_frameClass.validateFrame(e);
			try {
				_frameClass.addInstance(e);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone "
				+ "badly wrong for an InvalidFrameClass"
				+ " to be thrown here!");
			}
			e.setNumber(_frameClass.getInstances().size());
		}
	}

	/**********************************************************************/
	// LiteralFrame.buildFrame : void -> void
	/***********************************************************************
	** Fills in the slots for this frame from the underlying exjava
	** Literal object.
	***********************************************************************/

	public void buildFrame () {
		try {
			/////////////////////////////
			// literal_value : literal //
			/////////////////////////////
			Object value= null;
			switch (_underlyingLiteral.kind) {
				case Literal.INTEGER: value= new
				DefinedTypeInstance(DefinedType.lookup
				("integer_literal"),_underlyingLiteral.integer);
					break;
				case Literal.REAL: value= new
				DefinedTypeInstance(DefinedType.lookup
				("real_literal"),_underlyingLiteral.real);
					break;
				case Literal.LOGICAL: value= new
				DefinedTypeInstance(DefinedType.lookup
				("logical_literal"),_underlyingLiteral.logical);
					break;
				case Literal.STRING: value= new
				DefinedTypeInstance(DefinedType.lookup
				("string_literal"),_underlyingLiteral.string);
					break;
				case Literal.BINARY: value= new
				DefinedTypeInstance(DefinedType.lookup
				("integer_literal"),_underlyingLiteral.binary);
					break;
			}
			setSlotFiller("literal_value", value);
		} catch (Exception x) {
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "LiteralFrame.buildFrame!");
		}
	}

	/**********************************************************************/
	// LiteralFrame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}
	
	/**********************************************************************/
	// static LiteralFrame.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of
	** LiteralFrame.
	** @param oo An output object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClass);
		oo.writeObject(_cache);
	}

	/**********************************************************************/
	// static LiteralFrame.readClass : ObjectInput -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** LiteralFrame.
	** @param ois An input object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClass= (FrameClass)oi.readObject();
			_cache= (Hashtable)oi.readObject();
			_initialised= true;
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the "
			+ "stupid compiler is just too dumb to realise this!");
		}
	}
}
