     ////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM //
   // Written by Andrew Broad             ///
  //                                     ////
 // Slot.java (1998:10:27 11:36)        /////
/////////////////////////////////////////////
// Methods of Slot:                    //////
// - Slot                              //////
// - getSlotFrame                      //////
// - getName                           //////
// - getFrameClassDeclaredIn           //////
// - getType                           //////
// - isAggType                         //////
// - lookupInverseSlot                 /////
// - validFiller                       ////
// - static writeClass                 ///
// - static readClass                  //
////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `Slot' represents a slot in a frame. It has a name and a type.
** Each frame class has a set of slots associated with it, which instances of
** the frame class can have fillers for.
*******************************************************************************/

public class Slot implements Serializable {
	private Frame _slotFrame;
	private String _name;
	private FrameClass _frameClassDeclaredIn;
	private boolean _aggType;
	private static Hashtable/*String->
	               Hashtable<String->
	               Hashtable<FrameClass->Slot>>*/ _inverseSlots;

	/**********************************************************************/
	// Slot constructor : Attr ->
	/***********************************************************************
	** Constructs this slot by plugging in the given exjava Attr object
        ** which is taken to represent it (it does so by copying the reference,
        ** which is maximally efficient, but the client should not then alter
        ** the Attr directly, as that would compromise the integrity of the
        ** frame).
        ** @param slotFrame An exjava Attr object to wrap in this frame.
	** @param frameClass The frame class to which this slot belongs.
	***********************************************************************/

	public Slot (Attr slotFrame, FrameClass frameClass,
	             Express frameClassModel) {
		_frameClassDeclaredIn= frameClass;
		_name= slotFrame.id;
		_aggType= slotFrame.type instanceof AggType;

		////////////////////
		// Explicit slots //
		////////////////////
		if (slotFrame instanceof ExplicitAttr) {
			try {
				_slotFrame= new ExplicitAttrFrame
					((ExplicitAttr)slotFrame);
			} catch (FrameAlreadyExists y) {
				_slotFrame= y._frame;
			}
		} else
		///////////////////
		// Inverse slots //
		///////////////////
		if (slotFrame instanceof InverseAttr) {
			try {
				_slotFrame= new InverseAttrFrame
					((InverseAttr)slotFrame);
			} catch (FrameAlreadyExists y) {
				_slotFrame= y._frame;
			}

			if (_inverseSlots == null)
				_inverseSlots= new Hashtable();
			BaseType type= slotFrame.type;
			while (! (type instanceof NamedType)) {
				type= ((AggType)type).elemType;
			}
			String thatFrameClass= 
				((Entity)((NamedType)type).decl).id;
			String forAttrSlotName= 
				((InverseAttr)slotFrame).getForRef().id;
			String frameClassDeclaredIn= null;
			FrameClass thatFC= null;
			try {
				thatFC= FrameClass.lookup(thatFrameClass);
			} catch (FrameClassDoesNotExist w) {
				thatFC= new FrameClass
					((Entity)((NamedType)type).decl,
					frameClassModel);
			}
			try {
				frameClassDeclaredIn= thatFC.
					getSlot(forAttrSlotName).
					getFrameClassDeclaredIn().getName();
			} catch (InvalidSlot z) {
				System.err.println("Slot \"" + forAttrSlotName +
				"\" not found in FrameClass \"" + thatFrameClass
				+ "\"");
				System.exit(1);
			}
			Hashtable inverseSlots2= (Hashtable)
				_inverseSlots.get(frameClassDeclaredIn);
			if (inverseSlots2 == null) {
				inverseSlots2= new Hashtable();
				_inverseSlots.put(frameClassDeclaredIn,
				                  inverseSlots2);
			}
			Hashtable inverseSlots3= (Hashtable)
				inverseSlots2.get(forAttrSlotName);
			if (inverseSlots3 == null) {
				inverseSlots3= new Hashtable();
				inverseSlots2.put(forAttrSlotName,
				                  inverseSlots3);
			}
			inverseSlots3.put(frameClass,this);
		} else
		///////////////////
		// Derived slots //
		///////////////////
		if (slotFrame instanceof DerivedAttr) {
			System.err.println("Derived slots not supported yet!");
		} else {
			System.err.println("This print statement could not have"
				+ " been executed!!");
		}
	}

        /**********************************************************************/
        // Slot.getSlotFrame : void -> Frame
        /***********************************************************************
        ** Delivers up the underlying frame of this slot.
        ** @return The frame representing this slot.
        ***********************************************************************/
 
        public Frame getSlotFrame () {
                return _slotFrame;
        }

	/**********************************************************************/
	// Slot.getName : void -> String
	/***********************************************************************
	** Returns the name of this slot.
	** @return The name of this slot.
	***********************************************************************/
 	
	public String getName () {
		return _name;
	}
	
	/**********************************************************************/
	// Slot.getFrameClassDeclaredIn : void -> String
	/***********************************************************************
	** Returns the frame class in which this slot was declared (the
	** supertype from which it is inherited).
	** @return The name of this slot.
	***********************************************************************/
 	
	public FrameClass getFrameClassDeclaredIn () {
		return _frameClassDeclaredIn;
	}
	
	/**********************************************************************/
	// Slot.getType : void -> Object
	/***********************************************************************
	** Returns the type of this slot.
	** @return The type of this slot.
	***********************************************************************/
 
 	public Object getType () {
		try {
			return _slotFrame.getSlotFiller("tipe");
		} catch (InvalidSlot x) {
			System.err.println("Something has gone very badly "
				+ "wrong for Slot.getType to throw an "
				+ "InvalidSlot exception!");
			return null;
		}
	}

	/**********************************************************************/
	// Slot.isAggType : void -> boolean
	/***********************************************************************
	** Returns true just in case this slot is of an aggregation type (i.e.
	** to be filled by a set, bag, list or array).
	** @return Whether this slot is of an aggregation type.
	***********************************************************************/

	public boolean isAggType () {
		return _aggType;
	}

	/**********************************************************************/
	// Slot.lookupInverseSlot : (String x String x FrameClass) -> Slot
	/***********************************************************************
	** Looks up the inverse slot of a given slot in a given frame class
	** whose filler is frame of a given frame class.
	** @param frameClass The name of the frame class the slot whose inverse
	** is to be looked up is in.
	** @param slot The name of the slot whose inverse is to be looked up.
	** @param slotFillerFrameClass The frame class of the actual filler
	** (must be a frame) of `slot'.
	** @return The inverse slot of <i>frameClass</i>.<i>slot</i> for a
	** filler of type <i>slotFillerFrameClass</i>, or null if there isn't
	** one.
	***********************************************************************/

	public static Slot lookupInverseSlot (String frameClass, String slot,
	                               FrameClass slotFillerFrameClass)
	{
		if (_inverseSlots == null) return null;
		
		Hashtable inverseSlots2= (Hashtable)_inverseSlots.
		                                    get(frameClass);
		if (inverseSlots2 == null) return null;
		
		Hashtable inverseSlots3= (Hashtable)inverseSlots2.get(slot);
		if (inverseSlots3 == null) return null;

		Enumeration iS3Enum= inverseSlots3.keys();
		while (iS3Enum.hasMoreElements()) {
			FrameClass key= (FrameClass)iS3Enum.nextElement();
			if (key.subsumes(slotFillerFrameClass))
				return (Slot)inverseSlots3.get(key);
		}
		return null;
	}

	/**********************************************************************/
	// Slot.validFiller : Object -> boolean
	/***********************************************************************
	** Validates a potential slot filler. Returns true just in case the
	** given object is of the right type to fill this slot (or a subtype
	** thereof). Works either in the Java object sense (if the type of this
	** slot is specified as a Class object) or in the frame sense (if the
	** type of this slot is specified as a <a href="FrameClass.html">
	** FrameClass</a> object.
	** @param filler A candidate filler for this slot.
	** @return Whether `filler' could fill this slot.
	***********************************************************************/
 
	public boolean validFiller (Object filler) {
//System.err.println("Slot.validFiller called for " + getName());
return true;

/*
		if (getType() instanceof Class) {
			return ((Class)getType()).isInstance(filler);
		} else if (getType() instanceof FrameClass) {
			return ((FrameClass)getType()).subsumes(((Frame)filler).
				getFrameClass());
		} else {
				System.err.println("The type of slot "
					+ getName() + " is not either a Class "
					+ "or a FrameClass!!");
				return false;
		}
*/
	}

	/**********************************************************************/
	// static Slot.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of Slot.
	** @param oo An output stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_inverseSlots);
	}

	/**********************************************************************/
	// static Slot.readClass : ObjectInputStream -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of Slot.
	** @param oi An input stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_inverseSlots= (Hashtable)oi.readObject();
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the"
			+ "stupid compiler is just too dumb to realise this!");
		}
	}

	/**********************************************************************/
	// static Slot.showInverseSlots : void -> void
	/***********************************************************************
	** Shows the inverse slots lookup table - helpful for debugging.
	***********************************************************************/

	private static void showInverseSlots () {
		Enumeration inverseSlots= _inverseSlots.keys();
		System.err.println("++++++++++++++++++++++++++++++++++++++++" +
		                   "++++++++++++++++++++++++++++++++++++++++");
		System.err.println("Slot._inverseSlots");
		while (inverseSlots.hasMoreElements()) {
			String frameClass= (String)inverseSlots.nextElement();
			System.err.println("\t" + frameClass + " ->");
			Hashtable inverseSlots2= (Hashtable)_inverseSlots.
			                                    get(frameClass);
			Enumeration iS2= inverseSlots2.keys();
			while (iS2.hasMoreElements()) {
				String slot= (String)iS2.nextElement();
				System.err.println("\t\t" + slot + " ->");
				Hashtable inverseSlots3= (Hashtable)
				                        inverseSlots2.get(slot);
				Enumeration iS3= inverseSlots3.keys();
				while (iS3.hasMoreElements()) {
					FrameClass fc= (FrameClass)iS3.
					               nextElement();
					System.err.println("\t\t\t" + fc +
					" -> " + ((Slot)inverseSlots3.get(fc)).
					         getName());
				}
			}
		}
		System.err.println("++++++++++++++++++++++++++++++++++++++++" +
		                   "++++++++++++++++++++++++++++++++++++++++");
	}
}