     ////////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM     //
   // Written by Andrew Broad                 ///
  //                                         ////
 // UniqueRuleFrame.java (1998:10:27 11:38) /////
/////////////////////////////////////////////////
// UniqueRuleFrame is a subclass of Frame.  /////
//                                          /////
// Methods of UniqueRuleFrame:              /////
// - static initialise                      /////
// - UniqueRuleFrame(UniqueRule)            /////
// - UniqueRuleFrame(UniqueRule,Hashtable)  /////
// - static buildQueuedFrames               /////
// - buildFrame                             /////
// - getFrameClass                          /////
// - getNumber             [INHERITED]      /////
// - setNumber             [INHERITED]      /////
// - getSlotFiller(Slot)   [INHERITED]      /////
// - getSlotFiller(String) [INHERITED]      /////
// - setSlotFiller(Slot)   [INHERITED]      /////
// - setSlotFiller(String) [INHERITED]      /////
// - toString              [INHERITED]      /////
// - print                 [INHERITED]      /////
// - generateFIL           [INHERITED]      ////
// - static writeClass                      ///
// - static readClass                       //
/////////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `UniqueRuleFrame', a subclass of <a href="Frame.html">Frame</a>,
** represents a frame internally as an exjava UniqueRule object - such frames
** should only be used to represent unique rules in EXPRESS models.
*******************************************************************************/

public class UniqueRuleFrame extends Frame {
	private static FrameClass _frameClass;
	private static boolean _initialised= false;
	private UniqueRule _underlyingUniqueRule;
	private static Vector/*UniqueRuleFrame*/ _toBeBuilt;
	private static Hashtable/*UniqueRule,UniqueRuleFrame*/ _cache;
	
	/**********************************************************************/
	// static UniqueRuleFrame.initialise : Express -> void
	/***********************************************************************
	** Initialises the UniqueRuleFrame class by setting the class variable
	** _frameClass after looking up the `unique_rule' entity in the given
	** EXPRESS (meta)model.
	** @param metamodel An EXPRESS metamodel, which is expected to contain
	**  the entity `unique_rule'.
	***********************************************************************/

	public static void initialise (Express metamodel) {
		Entity e= metamodel.findEntity("unique_rule");
		if (e != null) {
			_frameClass= new FrameClass(e,metamodel);
		} else {
			System.err.println("Fatal error: failed to find "
			+ "ENTITY unique_rule in \"EXPRESS/"
			+ "frame_classes.exp\"!");
			System.exit(1);
		}
	}
	
	/**********************************************************************/
	// UniqueRuleFrame constructor : UniqueRule ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava UniqueRule
	** object (it does so by copying the reference, which is maximally
	** efficient, but the client should not then alter the UniqueRule
	** directly, as that would compromise the integrity of the frame).
	** @param instanceData An exjava UniqueRule object to wrap in this
	**                     frame.
	** @exception FrameAlreadyExists If an UniqueRuleFrame has already
	**  been constructed for `instanceData'.
	***********************************************************************/

	public UniqueRuleFrame (UniqueRule instanceData)
	                         throws FrameAlreadyExists
	{
		this(instanceData,new Hashtable());
	}
	
	/**********************************************************************/
	// UniqueRuleFrame constructor : (UniqueRule x Hashtable) ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava UniqueRule
	** object (it does so by copying the reference, which is maximally
	** efficient, but the client should not then alter the UniqueRule
	** directly, as that would compromise the integrity of the frame). This
	** constructor allows the unique rule frame to be tagged with
	** extra slots, given in a hash table.
	** @param instanceData An exjava UniqueRule object to wrap in this
	**                     frame.
	** @param moreInstanceData Extra slots to tag this unique rule frame
	**  with.
	** @exception FrameAlreadyExists If an UniqueRuleFrame has already
	**  been constructed for `instanceData'.
	***********************************************************************/

	public UniqueRuleFrame (UniqueRule instanceData,
	                        Hashtable moreInstanceData)
	                        throws FrameAlreadyExists
	{
		super(moreInstanceData);

		if (_cache == null)
			_cache= new Hashtable();			

		_underlyingUniqueRule= instanceData;

		////////////////////////////////////////////////////////
		// Check that an UniqueRuleFrame has not already been //
		// constructed for this UniqueRule.                   //
		////////////////////////////////////////////////////////		
		if (_cache.containsKey(_underlyingUniqueRule))
			throw new FrameAlreadyExists((Frame)_cache.
				get(_underlyingUniqueRule));

		_cache.put(_underlyingUniqueRule,this);

		////////////////////////////////////////////////////////////
		// Build the frame now, provided its frame class has been //
		// constructed, otherwise queue it to be built later.     //
		////////////////////////////////////////////////////////////
		if (_initialised) {
			buildFrame();
			_frameClass.validateFrame(this);
			try {
				_frameClass.addInstance(this);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
					+ "wrong for an InvalidFrameClass to be"
						+ " thrown here!");
			}
			setNumber(_frameClass.getInstances().size());
		} else {
			if (_toBeBuilt == null)
				_toBeBuilt= new Vector();
			_toBeBuilt.addElement(this);
		}
	}
	
	/**********************************************************************/
	// static UniqueRuleFrame.buildQueuedFrames : void -> void
	/***********************************************************************
	** Builds any UniqueRuleFrames that are queued to be built (those that
	** were constructed before the frame class was constructed). Their slots
	** are filled from the exjava objects with which they were constructed,
	** and they are then validated.
	***********************************************************************/

	public static void buildQueuedFrames () {
		_initialised= true;
		if (_toBeBuilt == null)
			return;

		Enumeration toBeBuilt= _toBeBuilt.elements();
		while (toBeBuilt.hasMoreElements()) {
			UniqueRuleFrame e= (UniqueRuleFrame)
			                   toBeBuilt.nextElement();
			e.buildFrame();
			_frameClass.validateFrame(e);
			try {
				_frameClass.addInstance(e);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone "
				+ "badly wrong for an InvalidFrameClass"
				+ " to be thrown here!");
			}
			e.setNumber(_frameClass.getInstances().size());
		}
	}

	/**********************************************************************/
	// UniqueRuleFrame.buildFrame : void -> void
	/***********************************************************************
	** Fills in the slots for this frame from the underlying exjava
	** UniqueRule object.
	***********************************************************************/

	public void buildFrame () {
		try {
			////////////////////
			// label : STRING //
			////////////////////
			setSlotFiller("label",_underlyingUniqueRule.id);
		
			/////////////////////////////
			// expression : expression //
			/////////////////////////////
			setSlotFiller("attributes", new
				UnorderedVector(_underlyingUniqueRule.refs));

		} catch (Exception x) {
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "UniqueRuleFrame.buildFrame!");
		}
	}

	/**********************************************************************/
	// UniqueRuleFrame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}
	
	/**********************************************************************/
	// static UniqueRuleFrame.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of
	** UniqueRuleFrame.
	** @param oo An output object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClass);
		oo.writeObject(_cache);
	}

	/**********************************************************************/
	// static UniqueRuleFrame.readClass : ObjectInput -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** UniqueRuleFrame.
	** @param ois An input object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClass= (FrameClass)oi.readObject();
			_cache= (Hashtable)oi.readObject();
			_initialised= true;
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the "
			+ "stupid compiler is just too dumb to realise this!");
		}
	}
}
