     /////////////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM          //
   // Written by Andrew Broad                      ///
  //                                              ////
 // UserFnCallFrame.java (1998:10:27 11:39)      /////
//////////////////////////////////////////////////////
// UserFnCallFrame is a subclass of Frame.       /////
//                                               /////
// Methods of UserFnCallFrame:                   /////
// - static initialise                           /////
// - UserFnCallFrame(UserFnCall,Frame)           /////
// - UserFnCallFrame(UserFnCall,Frame,Hashtable) /////
// - static buildQueuedFrames                    /////
// - buildFrame                                  /////
// - getFrameClass                               /////
// - getNumber             [INHERITED]           /////
// - setNumber             [INHERITED]           /////
// - getSlotFiller(Slot)   [INHERITED]           /////
// - getSlotFiller(String) [INHERITED]           /////
// - setSlotFiller(Slot)   [INHERITED]           /////
// - setSlotFiller(String) [INHERITED]           /////
// - toString              [INHERITED]           /////
// - print                 [INHERITED]           /////
// - generateFIL           [INHERITED]           ////
// - static writeClass                           ///
// - static readClass                            //
//////////////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `UserFnCallFrame', a subclass of <a href="Frame.html">Frame</a>,
** represents a frame internally as an exjava UserFnCall object - such frames
** should only be used to represent user function calls in EXPRESS models.
*******************************************************************************/

public class UserFnCallFrame extends Frame {
	private static FrameClass _frameClass;
	private static boolean _initialised= false;
	private UserFnCall _underlyingUserFnCall;
	private static Vector/*UserFnCallFrame*/ _toBeBuilt;
	private static Hashtable/*UserFnCall,UserFnCallFrame*/ _cache;
	private Frame _theEntityOrType;
	
	/**********************************************************************/
	// static UserFnCallFrame.initialise : Express -> void
	/***********************************************************************
	** Initialises the UserFnCallFrame class by setting the class
	** variable _frameClass after looking up the `user_function_call'
	** entity in the given EXPRESS (meta)model.
	** @param metamodel An EXPRESS metamodel, which is expected to contain
	**  the entity `user_function_call'.
	***********************************************************************/

	public static void initialise (Express metamodel) {
		Entity e= metamodel.findEntity("user_function_call");
		if (e != null) {
			_frameClass= new FrameClass(e,metamodel);
		} else {
			System.err.println("Fatal error: failed to find "
			+ "ENTITY user_function_call in \"EXPRESS/"
			+ "frame_classes.exp\"!");
			System.exit(1);
		}
	}
	
	/**********************************************************************/
	// UserFnCallFrame constructor : UserFnCall ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava UserFnCall object
	** (it does so by copying the reference, which is maximally efficient
	** but the client should not then alter the UserFnCall directly, as that
	** would compromise the integrity of the frame).
	** @param instanceData An exjava UserFnCall object to wrap in this frame.
	** @param theEntity The entity which contains this function call.
 	** @exception FrameAlreadyExists If an UserFnCall has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public UserFnCallFrame (UserFnCall instanceData, Frame theEntity)
	                         throws FrameAlreadyExists
	{
		this(instanceData,theEntity,new Hashtable());
	}
	
	/**********************************************************************/
	// UserFnCallFrame constructor : (UserFnCall x Hashtable) ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava UserFnCall object
	** (it does so by copying the reference, which is maximally efficient,
	** but the client should not then alter the UserFnCall directly, as that
	** would compromise the integrity of the frame). This constructor allows
	** the function call frame to be tagged with extra slots, given in a
	** hash table.
	** @param instanceData An exjava UserFnCall object to wrap in this frame.
	** @param theEntity The entity which contains this function call.
	** @param moreInstanceData Extra slots to tag this function call
	**  frame with.
	** @exception FrameAlreadyExists If an UserFnCall has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public UserFnCallFrame (UserFnCall instanceData, Frame theEntity,
	                     Hashtable moreInstanceData)
	                    throws FrameAlreadyExists
	{
		super(moreInstanceData);

		if (_cache == null)
			_cache= new Hashtable();			

		_underlyingUserFnCall= instanceData;
		_theEntityOrType= theEntity;

		////////////////////////////////////////////////////////
		// Check that an UserFnCallFrame has not already been //
		// constructed for this UserFnCall.                   //
		////////////////////////////////////////////////////////
		if (_cache.containsKey(_underlyingUserFnCall))
			throw new FrameAlreadyExists((Frame)_cache.
				get(_underlyingUserFnCall));

		_cache.put(_underlyingUserFnCall,this);

		////////////////////////////////////////////////////////////
		// Build the frame now, provided its frame class has been //
		// constructed, otherwise queue it to be built later.     //
		////////////////////////////////////////////////////////////
		if (_initialised) {
			buildFrame();
			_frameClass.validateFrame(this);
			try {
				_frameClass.addInstance(this);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
				+ "wrong for an InvalidFrameClass to be thrown "
				+ "here!");
			}
			setNumber(_frameClass.getInstances().size());
		} else {
			if (_toBeBuilt == null)
				_toBeBuilt= new Vector();
			_toBeBuilt.addElement(this);
		}
	}
	
	/**********************************************************************/
	// static UserFnCallFrame.buildQueuedFrames : void -> void
	/***********************************************************************
	** Builds any UserFnCall Frames that are queued to be built (those that
	** were constructed before the frame class was constructed). Their slots
	** are filled from the exjava objects with which they were constructed,
	** and they are then validated.
	***********************************************************************/

	public static void buildQueuedFrames () {
		_initialised= true;
		if (_toBeBuilt == null)
			return;

		Enumeration toBeBuilt= _toBeBuilt.elements();
		while (toBeBuilt.hasMoreElements()) {
			UserFnCallFrame e= (UserFnCallFrame)toBeBuilt.nextElement();
			e.buildFrame();
			_frameClass.validateFrame(e);
			try {
				_frameClass.addInstance(e);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone "
				+ "badly wrong for an InvalidFrameClass"
				+ " to be thrown here!");
			}
			e.setNumber(_frameClass.getInstances().size());
		}
	}

	/**********************************************************************/
	// UserFnCallFrame.buildFrame : void -> void
	/***********************************************************************
	** Fills in the slots for this frame from the underlying exjava UserFnCall
	** object.
	***********************************************************************/

	public void buildFrame () {
		try {
			//////////////////////////////
			// function_called : STRING //
			//////////////////////////////
			setSlotFiller("function_called", _underlyingUserFnCall.
			                                 fn.id);

			////////////////////////////////////////////////////
			// actual_parameters : LIST [0 : ?] OF expression //
			////////////////////////////////////////////////////
			setSlotFiller("actual_parameters", new OrderedVector
				(_underlyingUserFnCall.params,
				 _theEntityOrType));
		} catch (Exception x) {
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "UserFnCallFrame.buildFrame!");
		}
	}

	/**********************************************************************/
	// UserFnCallFrame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}
	
	/**********************************************************************/
	// static UserFnCallFrame.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of
	** UserFnCallFrame.
	** @param oo An output object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClass);
		oo.writeObject(_cache);
	}

	/**********************************************************************/
	// static UserFnCallFrame.readClass : ObjectInput -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** UserFnCallFrame.
	** @param ois An input object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClass= (FrameClass)oi.readObject();
			_cache= (Hashtable)oi.readObject();
			_initialised= true;
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the "
			+ "stupid compiler is just too dumb to realise this!");
		}
	}
}
